from Nexgram.errors import *
from Nexgram.types import *
import asyncio

class Dispatch:
  async def dispatch_update(self, update):
    log = self.log
    for gf in self.on_listeners:
      asyncio.create_task(gf(update))
    if update.get('message'):
      try:
        m = update.get('message')
        frm = m.get('from')
        ch = m.get('chat')
        from_user = User(
          frm['id'],
          frm['first_name'],
          username=frm.get('username'),
          is_bot=frm.get('is_bot'),
          is_self=frm['id'] == self.me.id,
        )
        chat = Chat(
          id=ch['id'],
          title=ch.get('title'),
          first_name=ch.get('first_name'),
          last_name=ch.get('last_name'),
          type=ch.get('type'),
          username=ch.get('username')
        )
        reply_to_message = None
            
        message = Message(
          client=self,
          id=m['message_id'],
          from_user=from_user,
          chat=chat,
          reply_to_message=reply_to_message,
          text=m.get('text')
        )
        
        for x in self.on_message_listeners:
          asyncio.create_task(self.call(self.on_message_listeners, x, self, message))
      except Exception as e:
        log.error(f"[DispatchUpdate] Line 44: {e}, message: {m}")


import logging
import inspect
import asyncio

log = logging.getLogger(__name__)

class Filter:
  def __init__(self, func):
    self.func = func
    self.is_async = inspect.iscoroutinefunction(func)
  async def __call__(self, *args, **kwargs):
    if self.is_async:
      return await self.func(*args, **kwargs)
    return self.func(*args, **kwargs)
  def __and__(self, other):
    async def combined(*args, **kwargs):
      r1 = await self(*args, **kwargs) if self.is_async else self(*args, **kwargs)
      r2 = await other(*args, **kwargs) if other.is_async else other(*args, **kwargs)
      return r1 and r2
    return Filter(combined)
  def __or__(self, other):
    async def combined(*args, **kwargs):
      r1 = await self(*args, **kwargs) if self.is_async else self(*args, **kwargs)
      r2 = await other(*args, **kwargs) if other.is_async else other(*args, **kwargs)
      return r1 or r2
    return Filter(combined)
  def __invert__(self):
    async def inverted(*args, **kwargs):
      r1 = await self(*args, **kwargs) if self.is_async else self(*args, **kwargs)
      return not r1
    return Filter(inverted)
    
def create(func):
  if isinstance(func, Filter):
    return func
  name = getattr(func, "__name__", "CustomFilter")
  return type(name, (Filter,), {"__call__": func})(func)
     
text = create(lambda _, message: message.text)

def command(cmd, prefix=['/']):
  async def wrapper(_, __, m):
    p = next((p for p in prefix if m.text.startswith(p)), None)
    return p and (m.text[len(p):] in cmd if isinstance(cmd, list) else m.text[len(p):] == cmd)
  return create(wrapper)
  
def user(id):
  async def wrapper(_, __, m):
    if isinstance(id, (int, str)) and str(id).isdigit():
      return m.from_user.id == int(id)
    if isinstance(id, list):
      return any(user(_, __, m) for user in map(user, id))
    urls = ["http://t.me/", "https://t.me/", "www.t.me/", "@", "http://telegram.dog/", "https://telegram.dog/"]
    return any(id.replace(x, "") == m.from_user.username for x in urls)
  return create(wrapper)
  
  


"""Api class used to post/get from BotApi, I don't like to create code for calling api each time."""
import aiohttp

class Api:
  async def post(self, url, json: dict = None, need_text=False):
    async with aiohttp.ClientSession() as mano:
      async with mano.post(url=url, json=json) as ily:
        if need_text:
          return await ily.text()
        else: return await ily.json()

  async def get(self, url, params=None, need_text=False):
    async with aiohttp.ClientSession() as mano:
      async with mano.get(url=url, params=params) as ily:
        if need_text:
          return await ily.text()
        else: return await ily.json()


from Nexgram.errors import BadRequest
import asyncio

class CopyMessages:
  async def copy_messages(self, chat_id: int, from_chat_id: int, id, caption=None, parse_mode=None):
    if not self.connected: raise ConnectionError("Client is not connected, you must connect the client to copy message.")
    if isinstance(id, int): id = [id]
    output = []
    for x in id:
      if not isinstance(x, int) or not str(x).isdigit():
        raise TypeError("You should only pass integer in message_id")
      if output: asyncio.sleep(0.200)
      data = {
        "chat_id": chat_id,
        "from_chat_id": from_chat_id,
        "message_id": int(x)
      }
      if caption: data['caption'] = caption
      if parse_mode: pass
      z = await self.api.post(self.ApiUrl+"copyMessage", json=data)
      if not z.get('ok') and z.get('error_code'):
        error_type = z.get('description')
        error = z.get('description').split(':', 1)[1]
        if 'bad request' in error_type.lower():
          raise BadRequest(error)
      output.append(z)
    return output


from .on_disconnect import OnDisconnect
from .on import On
from .on_message import OnMessage
from .on_callback_query import OnCallbackQuery
from .on_inline_query import OnInlineQuery

class Decorators(
  On,
  OnDisconnect,
  OnMessage,
  OnCallbackQuery,
  OnInlineQuery,
):
  pass


from .send_message import sendMessage
from .dispatch_update import Dispatch
from .decorators import Decorators
from .call import Call
from .forward_messages import ForwardMessages
from .copy_messages import CopyMessages

class Methods(
  sendMessage,
  Dispatch,
  Decorators,
  Call,
  ForwardMessages,
  CopyMessages,
):
  pass


import json
from Nexgram.errors import *

class Message:
  def __init__(
    self,
    client: "Nexgram.Client",
    id: int,
    from_user: "Nexgram.types.User",
    chat: "Nexgram.types.Chat",
    reply_to_message: "Nexgram.types.Message",
    text: str = None,
  ):
    from Nexgram.types import User, Chat
    from Nexgram import Client
    
    if not isinstance(from_user, User): raise InvalidObject("You should pass 'Nexgram.types.User' object in 'from_user' argument not others.")
    if not isinstance(chat, Chat): raise InvalidObject("You should pass 'Nexgram.types.Chat' object in 'chat' argument not others.")
    if not isinstance(client, Client): raise InvalidObject("You should pass 'Nexgram.Client' object in 'client' argument not others")
    
    self._ = "Nexgram.types.Message"
    self.id = id
    self.from_user = from_user
    self.chat = chat
    if reply_to_message:
      if not isinstance(reply_to_message, self):
        raise InvalidObject("You should pass 'Nexgram.Client.Message' object in 'reply_to_message' argument not others")
      self.reply_to_message = reply_to_message
    self.text = text
    self.client = client
  
  def __repr__(self):
    mf = ["client"]
    data = {k: v for k, v in self.__dict__.items() if k not in mf}
    return json.dumps(
      data,
      indent=2,
      ensure_ascii=False,
      default=lambda o: o.__dict__ if hasattr(o, "__dict__") else o
    )
    
  async def reply(self, text: str, reply_markup = None,parse_mode: str = None):
    client = self.client
    await client.send_message(
      chat_id=self.chat.id,
      text=text,
      reply_markup=reply_markup,
      reply_to_message_id=self.id,
      parse_mode=parse_mode,
    )
  async def delete(self):
    client, api, url = self.client, self.client.api, self.client.ApiUrl
    return await api.post(url+"deleteMessage", {"chat_id": self.chat.id, "message_id": self.id})
  async def forward(self, chat_id):
    client, api, url = self.client, self.client.api, self.client.ApiUrl
    return await client.forward_messages(chat_id, self.chat.id, self.id)
  async def copy(self, chat_id, caption=None, parse_mode=None):
    client, api, url = self.client, self.client.api, self.client.ApiUrl
    return await client.copy_messages(chat_id, self.chat.id, self.id, caption=caption, parse_mode=parse_mode)



import json
from typing import List
from .inline_keyboard_button import InlineKeyboardButton

class InlineKeyboardMarkup:
  def __init__(self, inline_keyboard: List[List["types.InlineKeyboardButton"]]):
    self._ = "Nexgram.types.InlineKeyboardMarkup"
    if not isinstance(inline_keyboard, list):
      raise TypeError("Failed to read buttons, you should pass list always!")
    
    for x in inline_keyboard:
      if not isinstance(x, list):
        raise TypeError("Failed to read buttons, you should pass 2 list always!")
      for z in x:
        if not isinstance(z, InlineKeyboardButton):
          raise TypeError("Failed to read buttons, you should always pass 'Nexgram.types.InlineKeyboardButton' object always!")
    
    self.inline_keyboard = inline_keyboard

  def __repr__(self):
    data = {k: v for k, v in self.__dict__.items()}
    return json.dumps(
      data,
      indent=2,
      ensure_ascii=False,
      default=lambda o: o.__dict__ if hasattr(o, "__dict__") else o
    )

  def read(self):
    return json.dumps({
      "inline_keyboard": [
        [{k: v for k, v in button.__dict__.items() if not k.startswith("_")} for button in row]
        for row in self.inline_keyboard
      ]
    })


from Nexgram.filters import Filter
from Nexgram import filters as f

class OnInlineQuery:
  def on_inline_query(self, filters):
    if not isinstance(filters, Filter) and not filters is None:
      filters = f.create(filters)  

    def decorator(mano):
      if mano in self.on_inline_query_listeners:
        raise Exception("You have already used this same decorator, you cannot use it multiple times!")
      self.on_inline_query_listeners[mano] = filters if isinstance(filters, Filter) else None
      return mano
    return decorator


import asyncio
import inspect

class Call:
  async def call(self, src, func, *args):
    if func not in src:
      return
    filter_func = src.get(func)
    if filter_func:
      if inspect.iscoroutinefunction(filter_func.func):
        passed = await filter_func(*args)  
      else:
        passed = await asyncio.to_thread(filter_func.func, *args)
      if passed:
        await func(*args)
    else:
      await func(*args)


import json

class User:
  def __init__(
    self,
    id: int,
    first_name: str,
    last_name: str = None,
    username: str = None,
    is_self: bool = False,
    is_bot: bool = False,
  ):
    self._ = "Nexgram.types.User"
    self.id = id
    self.is_self = is_self
    self.is_bot = is_bot
    self.first_name = first_name
    if last_name: self.last_name = last_name
    if username: self.username = username
      
  def __repr__(self):
    return json.dumps(self.__dict__, indent=2, ensure_ascii=False)


import json

class InlineKeyboardButton:
  def __init__(
    self,
    text,
    url=None,
    callback_data=None,
    callback_game=None,
    user_id: int = None,
    switch_inline_query=None,
    switch_inline_query_current_chat=None,
  ):
    self._ = "Nexgram.types.InlineKeyboardButton"
    self.text = text
    if url:
      self.url = url
    elif callback_data:
      self.callback_data = callback_data
    elif callback_game:
      self.callback_game = callback_game
    elif user_id:
      self.url = f"tg://user?id={user_id}"
    elif switch_inline_query:
      self.switch_inline_query = switch_inline_query
    elif switch_inline_query_current_chat:
      self.switch_inline_query_current_chat = switch_inline_query_current_chat
    else: raise Exception("?.")
  def __repr__(self):
    return json.dumps(self.__dict__, indent=2, ensure_ascii=False)
    


import logging
import httpx
import aiohttp
import asyncio
from .methods import *
from .errors import *
from .types import *
from .api import Api

log = logging.getLogger(__name__)
clients = []

class Client(Methods):
  def __init__(
    self,
    name: str,
    bot_token: str,
  ):
    self.name = name
    self.bot_token = bot_token
    self.connected = False
    self.me = None
    self.offset = 0
    self.polling = False
    self.ApiUrl = f"https://api.telegram.org/bot{self.bot_token}/"
    self.api = Api()
    self.log = log
    clients.append(self)
    # Decorators --
    self.on_message_listeners = {}
    self.on_disconnect_listeners = {}
    self.on_callback_query_listeners = {}
    self.on_inline_query_listeners = {}
    self.on_listeners = {}
    
  async def start(self, start_polling=False):
    url = f"https://api.telegram.org/bot{self.bot_token}/getMe"
    async with aiohttp.ClientSession() as session:
      async with session.get(url) as r:
        r = await r.json()
        if r.get("ok"):
          self.connected = True
          r = r["result"]
          self.me = User(
            r['id'],
            r['first_name'],
            username=r['username'],
            is_self=True,
            is_bot=True,
          )
          log.info(f"Client connected as {self.me.first_name} (@{self.me.username})")
          if start_polling:
            try:
              loop = asyncio.get_running_loop()
            except RuntimeError:
              loop = asyncio.new_event_loop()
              asyncio.set_event_loop(loop)
            loop.create_task(self.start_polling())
            log.info("Exp. Feature Started: Loop created.")
            self.loop = loop
          return self.me
        raise ValueError("Failed to connect with your bot token. Please make sure your bot token is correct.")

  async def start_polling(self):
    if not self.connected:
      raise ConnectionError("Client is not connected. Please connect the client and start polling.")
    elif self.polling: raise PollingAlreadyStartedError("Polling already started, why you trying again and again? didn't you receive any updates?")
    self.polling = True
    log.info("Nexgram.py - polling started!")
    first_start = True
    max_retry, retry = 25, 0
    while self.polling:
      try:
        params = {"offset": self.offset, "timeout": 15}
        updates = await self.api.get(self.ApiUrl+"getUpdates", params=params)
        if "result" in updates and not first_start:
          for update in updates["result"]:
            self.offset = update["update_id"] + 1
            asyncio.create_task(self.dispatch_update(update))
        elif "result" in updates and first_start:
          first_start = False
      except Exception as e:
        if retry > max_retry:
          break
        log.error(f"[{retry}/{max_retry}] Error in polling: {e}")
        retry += 1
    await self.stop()
  
  async def stop(self):
    await self.trigger_disconnect()
    self.polling = False
    self.connected = False
    log.info("Client stopped.")


import aiohttp
from Nexgram.errors import BadRequest
from Nexgram.types import InlineKeyboardMarkup

class sendMessage:
  async def send_message(
    self,
    chat_id,
    text,
    reply_markup: "Nexgram.types.InlineKeyboardMarkup" = None,
    reply_to_message_id: int = None,
    parse_mode=None,
  ):
    if not self.connected: raise ConnectionError("Client is not connected, you must connect the client to send message.")
    url = f"https://api.telegram.org/bot{self.bot_token}/sendMessage"
    data = {
      "chat_id": chat_id,
      "text": text,
    }
    
    if parse_mode:
      data["parse_mode"] = parse_mode
    if reply_to_message_id:
      data["reply_to_message_id"] = reply_to_message_id
    if reply_markup:
      if not isinstance(reply_markup, InlineKeyboardMarkup):
        raise TypeError("You should pass 'Nexgram.types.InlineKeyboardMarkup' in reply_markup not others!")
      data['reply_markup'] = reply_markup.read()
    z = await self.api.post(url, json=data)
    if not z.get('ok') and z.get('error_code'):
      error_type = z.get('description')
      error = z.get('description').split(':', 1)[1]
      if 'bad request' in error_type.lower():
        raise BadRequest(error)
    return z


import json

class Chat:
  def __init__(
    self,
    id: int,
    title: str = None,
    first_name: str = None,
    last_name: str = None,
    type: str = None,
    username: str = None,
  ):
    self._ = "Nexgram.types.Chat"
    self.id = id
    if not title:
      if not first_name: 
        raise Exception("You didn't passed title or first_name, you should pass any of them.")
      self.first_name = first_name
      self.last_name = last_name
    else: self.title = title
    self.type = type
    if username: self.username = username
  
  def __repr__(self):
    return json.dumps(self.__dict__, indent=2, ensure_ascii=False)


from setuptools import setup, find_packages
import re

with open("requirements.txt", encoding="utf-8") as mano:
  requires = [z.strip() for z in mano]
    
with open("Nexgram/__init__.py", encoding="utf-8") as fk:
  version = re.findall(r"__version__ = \"(.+)\"", fk.read())[0]

setup( 
  name="Nexgram.py",
  version=version,
  packages=find_packages(),
  install_requires=requires,
  author="Otazuki",
  author_email="otazuki004@gmail.com",
  description="just a try",
  long_description=open("README.md").read(),
  long_description_content_type="text/markdown",
  url="https://github.com/Otazuki004/Nexgram.py",
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
  python_requires='>=3.1',
)


import logging
import signal
from signal import signal as signal_fn, SIGINT, SIGTERM, SIGABRT
from .client import clients
import asyncio

log = logging.getLogger(__name__)

signals = {
    k: v for v, k in signal.__dict__.items()
    if v.startswith("SIG") and not v.startswith("SIG_")
}

async def idle():
  task = None  
  def signal_handler(signum, __):
    log.info(f"Stop signal received ({signals[signum]}). Exiting...")
    task.cancel()

  for s in (SIGINT, SIGTERM, SIGABRT):
    signal_fn(s, signal_handler)
  while True:
    task = asyncio.create_task(asyncio.sleep(600))
    try: await task
    except asyncio.CancelledError:
      for x in clients:
        await x.stop()
      break


import asyncio

class OnDisconnect:
  def on_disconnect(self, mano):
    self.on_disconnect_listeners[mano] = None
  async def trigger_disconnect(self):
    for mano in self.on_disconnect_listeners:
      asyncio.create_task(mano(self))